c-----------------------------------------------------------------------
      program calrank
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912              MAIN
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read file of modeled data from a CALPUFF-type data file
c               and rank all values in magnitude for each species and
c               receptor.  Indentify the nth-highest values and the
c               percentile values requested by the user (each receptor
c               and species) and tabulate largest.  Create plot-files
c               for each requested rank/percentile.
c
c --- UPDATES:
c --- v1.2 (131113) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c           MODIFIED: HD_CFG, HD_PROC
c                     MODULES - HEADER
c
c --- v1.1 (120915) to v1.2 (131113)
c        1. Change definition of the percentile to be consistent
c           with USEPA tabulations of percentiles and ranks
c           P(rank N of M values)=100.(1.-N/M)
c           N(percentile P of M values)=M(1.-P/100.) {round-up}
c             [e.g. N=7.01 rounds up to N=8]
c           MODIFIED: COMP2
c
c --- v1.0 (120630) to v1.1 (120915)
c        1. Change all file pathname variables from C*132 to C*256
c           and make control file line-input strings C*320
c           MODIFIED: (main), CONTROL, OUTPUT
c                     MODULES v1.1 - USERINF
c                     UTILS   v1.1 - GETCMD
c        2. Output filename for percentiles used f4.1 for percentile
c           which fails for 99.95 (or larger) percentile:  Change to
c           f6.3 format
c           MODIFIED: OUTPUT
c        3. Format for output percentile values changed from 1 decimal
c           to 3 decimal places
c           MODIFIED: OUTPUT, XYCOUT
c        4. Add user-selected mass units for output
c           MODIFIED: CONTROL, OUTPUT, XYCOUT, ALLOC_2, DEALLOC_2
c                     MODULES v1.1 - USERINF, HEADER
c
c  EXTERNAL ROUTINES:   GETCMD, CONTROL, ALLOC_1, ALLOC_2, ALLOC_3,
c                       HD_CFG, HD_PROC, MK_DAFILE, COMP1, COMP2,
c                       OUTPUT, DEALLOC_1, DEALLOC_2
c-----------------------------------------------------------------------
c --- Access module data
      use inpdata
      use userinf
      use rankdata

      implicit none

      integer :: ictrl,iolst,ioinp,ioda,ioout
      parameter(ictrl=10, iolst=11, ioinp=13, ioout=14, ioda=15)

      character(len=12)   :: cversion,clevel
      character(len=256)  :: cfname

      integer :: ierr, ione
      data ione/1/

c --- Identify version and level of CALRANK:
      cversion='TNG-2.0.0'
      clevel='140912'

c --- Control file name (default or specified on command line)
      cfname='CALRANK.INP'
      call GETCMD(cfname)

c --- Open and process control file (opens list file also)
      open(ictrl,file=cfname,status='old')
      call CONTROL(ictrl,iolst,cversion,clevel)
      close(ictrl)

c --- Open file of model output (all receptors, species, times)
      open(ioinp,file=datfile,status='old',
     &           form='unformatted')

c --- Allocate arrays that depend only on the number of
c --- concurrent CALPUFF files
      call ALLOC_1(ione)

c --- Process top of file header for model run configuration
      call HD_CFG(iolst,ioinp,ione,cversion,clevel,ierr)
      if(ierr.NE.0) stop 'HALTED at call to HD_CFG -- see list file'

c --- Establish Direct Access file for storing image of data
      call MK_DAFILE(iolst,ioda,ione,ierr)
      if(ierr.NE.0) stop 'HALTED at call to MK_DAFILE -- see list file'

c --- Allocate remaining arrays for constructing DA file
      call ALLOC_2(ione)

c --- Process rest of file header
      call HD_PROC(ioinp,ione)

c --- Acquire data and store into DA file
      call COMP1(iolst,ioinp,ioda,ierr)
      if(ierr.NE.0) stop 'HALTED at call to COMP1 --- see list file'

c --- Release space for reading data file
      call DEALLOC_1

c --- Allocate arrays for processing DA file
      call ALLOC_3(ione)

c --- Process data in DA file to obtain requested ranks
      call COMP2(iolst,ioda,ierr)
      if(ierr.NE.0) stop 'HALTED at call to COMP2 --- see list file'

c --- Write results to output files
      call OUTPUT(iolst,ioout)

c --- Release space
      call DEALLOC_2

c --- Close files
      CLOSE(ioinp)
      CLOSE(iolst)
 
      write(*,*)
      write(*,*)'CALRANK application completed successfully'

      stop
      end

c---------------------------
c --- Add utility files
      include 'utils.for'
      include 'slatec.for'
c---------------------------


c-----------------------------------------------------------------------
      subroutine control(ictrl,iolst,cversion,clevel)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120915           CONTROL
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Obtain program control information
c               Open list file and write header
c
c --- UPDATES:
c --- v1.0 (120630) to v1.1 (120915)
c        1. Change all file pathname variables from C*132 to C*256
c           and make all line-input strings C*320
c        2. Add user-selected mass units for output
c
c --- INPUT:
c               ICTRL - integer   - Unit number of control file
c               IOLST - integer   - Unit number of output list file
c
c --- OUTPUT:
c     Module /USERINF/
c
c --- CONTROL called by: (main)
c --- CONTROL calls:     DELIM, MK_CASE, RM_BLANK
c----------------------------------------------------------------------
c --- Access module data
      use inpdata
      use userinf
      use rankdata

      implicit none

c --- Declare arguments
      integer :: ictrl,iolst
      character(len=*)  ::  cversion,clevel

c --- Locals
      logical :: lprocess
      integer :: i,ieof,ierr, icase
      integer :: ipct
      integer :: id1,id2,ieq
      integer :: l1,l2,r1,r2, num
      character(len=320)  ::  blank320,aline320
      character(len=5)    ::  ulcase(2)
      character(len=1)    ::  dlim

      data dlim/'!'/
      data ulcase/'UPPER','LOWER'/

c --- Establish initial values (others in module userinf)
      do i=1,320
         blank320(i:i)=' '
      enddo
      lstfile=blank320(1:256)
      lstfile(1:10)='calrank.lst'


c --- Scan control file for the list-file name, and
c --- number of reporting ranks (must allocate storage arrays)
c -----------------------------------------------------------
c --- (List file is not opened yet so error-writes will go to
c --- a FORT-file)
      n_nth=0
      n_pct=0
      ieof=0
      do while (ieof.EQ.0)
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Set range for variable name on left and value on right
            l1=id1+1
            l2=ieq-1
            r1=ieq+1
            r2=id2-1
            num=r2-r1+1
c ---       Convert to upper case
            call MK_CASE(iolst,'UPPER',aline320,ierr)
            if(ierr.NE.0) stop 'HALTED in CONTROL'
c ---       Interpret subset of inputs
            if(INDEX(aline320(l1:l2),'NTH_HIGHEST').GT.0) then
               n_nth=n_nth+1
            elseif(INDEX(aline320(l1:l2),'PERCENTILE').GT.0) then
               n_pct=n_pct+1
            elseif(INDEX(aline320(l1:l2),'LSTFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               lstfile=blank320(1:256)
               lstfile=aline320(r1:r2)
c ---          Left-justify
               lstfile=ADJUSTL(lstfile)
            elseif(INDEX(aline320(l1:l2),'LCFILES').GT.0) then
               if(INDEX(aline320(r1:r2),'T').GT.0) then
                  LCFILES=.TRUE.
               else
                  LCFILES=.FALSE.
               endif
            endif
         endif
      enddo

c --- Sum the number of requested ranks
      n_ranks=n_nth+n_pct

c --- Set pointer for ULCASE (1=upper, 2=lower)
      icase=1
      if(lcfiles) icase=2

c --- Open list file and write header
      call MK_CASE(iolst,ulcase(icase),lstfile,ierr)
      if(ierr.NE.0) stop 'HALTED in CONTROL'
      open(iolst,file=lstfile,status='unknown')
      write(iolst,*)
      write(iolst,*)
      write(iolst,*)'                    '//
     &              '    CALRANK RESULTS SUMMARY'
      write(iolst,*)'                   VERSION:'//cversion//
     &              '   LEVEL:'//clevel
      write(iolst,*)
      write(iolst,*)

c --- Rewind control file to complete the processing
      REWIND(ictrl)

c -------------------------------------------------------------
c --- Allocate arrays for storing ranks
      ALLOCATE (nth(n_nth),nth_from_pct(n_pct))
      ALLOCATE (pct(n_pct),pct_from_nth(n_pct))
c -------------------------------------------------------------

c --- Process control file again for all contents
c ------------------------------------------------
      n_nth=0
      n_pct=0
      ieof=0
      do while (ieof.EQ.0)
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Set range for variable name on left and value on right
            l1=id1+1
            l2=ieq-1
            r1=ieq+1
            r2=id2-1
            num=r2-r1+1
c ---       Convert to upper case
            call MK_CASE(iolst,'UPPER',aline320,ierr)
            if(ierr.NE.0) stop 'HALTED in CONTROL -- see list file'
c ---       Check list of expected inputs
c ---       -----------------------------

            if(INDEX(aline320(l1:l2),'DATFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               datfile=blank320(1:256)
               datfile=aline320(r1:r2)
               datfile=ADJUSTL(datfile)
               call MK_CASE(iolst,ulcase(icase),datfile,ierr)
               if(ierr.NE.0) stop 'HALTED in CONTROL -- see list file'

            elseif(INDEX(aline320(l1:l2),'LSTFILE').GT.0) then
c ---          Nothing:  already processed

            elseif(INDEX(aline320(l1:l2),'LCFILES').GT.0) then
c ---          Nothing:  already processed

            else
c ---          Remove blanks and reset pointers for rest
               call RM_BLANK(aline320)
               call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
c ---          Set range for variable name on left and value on right
               l1=id1+1
               l2=ieq-1
               r1=ieq+1
               r2=id2-1

               if(INDEX(aline320(l1:l2),'ICDAY').GT.0) then
                  read(aline320(r1:r2),'(i)') icday

               elseif(INDEX(aline320(l1:l2),'MASS_UNIT').GT.0) then
                  read(aline320(r1:r2),'(i)') mass_u

               elseif(INDEX(aline320(l1:l2),'NTH_HIGHEST').GT.0) then
                  n_nth=n_nth+1
                  read(aline320(r1:r2),'(i)') nth(n_nth)

               elseif(INDEX(aline320(l1:l2),'PERCENTILE').GT.0) then
                  n_pct=n_pct+1
                  if(INDEX(aline320(r1:r2),'.').EQ.0) then
                     read(aline320(r1:r2),'(i)') ipct
                     pct(n_pct)=FLOAT(ipct)
                  else
                     read(aline320(r1:r2),'(f)') pct(n_pct)
                  endif

               else
                  write(iolst,*)'Error in CALRANK control file'
                  write(iolst,*)'Assignment variable unknown:'
                  write(iolst,*)TRIM(aline320)
                  stop 'HALTED in CONTROL -- see list file'
               endif
            endif
         endif
      enddo

c --- Announce file that is processed to list file
      write(iolst,*)'File Processed: '//TRIM(datfile)
      write(iolst,*)

c --- Validate mass units selector
      if(mass_u.LT.1 .OR. mass_u.GT.5) then
         write(iolst,*)
         write(iolst,*)'Invalid output mass units selection: ',mass_u
         write(iolst,*)'Expected 1 - 5'
         ierr=ierr+1
      endif

c --- End it?
      if(ierr.GT.0)then
         stop 'HALTED in CONTROL -- see list file'
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine mk_dafile(iolst,ioda,k,ierr)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630         MK_DAFILE
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Define and open a Direct Access scratch file for
c               storing all of the data from the input file for
c               subsequent processing
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c           IODA - integer - File unit number for scratch DA file
c              k - integer - Sequential number for input CALPUFF file
c                            (must be 1 at present)
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- MK_DAFILE called by: (main)
c --- MK_DAFILE calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header
      use daccess

      implicit none

c --- Declare arguments
      integer :: iolst,ioda,k,ierr

c --- Locals
      integer :: nrecbytes

c --- QA arguments
      if(k.NE.1) then
         write(iolst,*)'ERROR in MK_DAFILE -- Invalid argument k: ',k
         write(iolst,*)'Expected just 1 CALPUFF file'
         ierr=1
         return
      endif

c --- Store data for 1 day (86400 seconds) in 1 DA record (plus MAX),
c --- for 1 species, for 1 receptor
      nvarrcrd=86400/(iavg(k)*nsecdt(k))+1
      if((86400/((nvarrcrd-1)*nsecdt(k))).NE.iavg(k)) then
         write(iolst,*)'ERROR in MK_DAFILE -- Problem with times '
         write(iolst,*)'Computed output times/day : ',(nvarrcrd-1)
         write(iolst,*)'for model timesteps (sec) : ',nsecdt(k)
         write(iolst,*)'and interval (timesteps)  : ',iavg(k)
         ierr=1
         return
      endif

c --- Open Direct Access scratch file 
c --- Record length in bytes holds 6 integers and 11 real arrays
c --- dimensioned MXRISE
      nrecbytes=4*nvarrcrd
      OPEN(ioda,access='direct',form='unformatted',recl=nrecbytes,
     &     status='scratch')

c --- Number of DA records needed for 1 day
c --- (date-time record plus nreceptors*nspecies)
      nrcrdday=1+ntotrec(k)*nspout(k)

      return
      end

c-----------------------------------------------------------------------
      subroutine hd_cfg(iolst,ioinp,k,cversion,clevel,ierr)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912            HD_CFG
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read first section of a CALPUFF output file header to
c               identify model configuration
c               Write comment records to header of results list file
c
c     NOTE:     The input file conforms to CALPUFF CONC.DAT format
c               unless it contains RUNNING AVERAGES.  This is indicated
c               within the first 6 COMMENT records.
c
c --- UPDATES:
c
c --- v1.0 (120630) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input CALPUFF file
c              k - integer - Sequential number for input CALPUFF file
c       CVERSION - char*12 - CALRANK version
c         CLEVEL - char*12 - CALRANK level (e.g., date)
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- HD_CFG called by: (main)
c --- HD_CFG calls:     UTC_ZONE
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header

      implicit none

c --- Arguments
      integer :: iolst,ioinp,k,ierr
      character(len=12)   :: cversion,clevel

c --- Locals
      integer :: msource_out
      integer :: i,iok,idum
      character(len=132)  :: comment, blank132

c --- Use variable names as listed in the CALPUFF Users Guide

c --- Initializations
      lrunning(k)=.FALSE.
      ierr=0
      idum=0
      do i=1,132
         blank132(i:i)=' '
      enddo
      msource_out=0

c --- Record #1 - File Declaration -- 24 words
      read(ioinp)  dataset(k),dataver(k),datamod(k)

c --- QA dataset names
      iok=0
      iok=iok+INDEX(dataset(k),'CONC.DAT')
      iok=iok+INDEX(dataset(k),'DFLX.DAT')
      iok=iok+INDEX(dataset(k),'WFLX.DAT')
      iok=iok+INDEX(dataset(k),'TFLX.DAT')
      if(iok.EQ.0) then
         write(iolst,*)'ERROR in HD_CFG -- Invalid dataset: '//
     &                  dataset//dataver
         write(iolst,*)'Requires Dataset version 2.1 and later format'
         ierr=1
         return
      endif

c --- Comment records
      read(ioinp)  ncom(k)

c --- Some comments are written to list file
      write(iolst,*)'Selected Header Information From File:'
c --- Marker line
      do i=1,132
         comment(i:i)='-'
      enddo
      write(iolst,'(a132)') comment

      do i=1,ncom(k)
         comment=blank132
         read(ioinp)  comment

c ---    Write the first 6 comment records to list file
c ---    and test for running averages
         if(i.LE.6) then
            if(INDEX(comment,'RUNNING').GT.0) lrunning(k)=.TRUE.
            write(iolst,'(a)') TRIM(comment)
         endif
      enddo

c --- Marker line
      do i=1,132
         comment(i:i)='-'
      enddo
      write(iolst,'(a132)') comment

c --- Control variables (record NCOM+3)

      if(dataver(k).EQ.'2.1             ') then
c ---------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),xbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
         nsrctype(k)=8
c ---    Rest of source-types are zero;  8 in this file
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    No receptor groups
         nrgrp(k)=0
c ---    No specie-level units in header
         lspunit(k)=.FALSE.
c ---    Translate real time zone to character
         call UTC_ZONE(iolst,'REAL2CHR',axbtz(k),xbtz(k),idum,ierr)
         if(ierr.NE.0) stop 'ERROR in HD_CFG -- See list file'

      elseif(dataver(k).EQ.'2.2             ') then
c -------------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),axbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
         nsrctype(k)=8
c ---    Rest of source-types are zero;  8 in this file
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    No receptor groups
         nrgrp(k)=0
c ---    Specie-level units are in header
         lspunit=.TRUE.
c ---    Translate character time zone to real
         call UTC_ZONE(iolst,'CHR2REAL',axbtz(k),xbtz(k),idum,ierr)
         if(ierr.NE.0) stop 'ERROR in HD_CFG -- See list file'

      elseif(dataver(k).EQ.'TNG-3.0         ') then
c ---------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),axbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),nsrctype(k),msource(k),
     &      ndrec(k),nrgrp(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
c ---    Number of each type of source in next record (record NCOM+3a)
         read(ioinp) (nsbytype(i,k),i=1,nsrctype(k))
c ---    Rest of source-types are zero
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    Specie-level units are in header
         lspunit=.TRUE.
c ---    Translate character time zone to real
         call UTC_ZONE(iolst,'CHR2REAL',axbtz(k),xbtz(k),idum,ierr)
         if(ierr.NE.0) stop 'ERROR in HD_CFG -- See list file'

      else
         write(iolst,*)'ERROR in HD_CFG -- Invalid dataset: '//
     &                  dataset//dataver
         write(iolst,*)'Requires Dataset version 2.1 and later format'
         ierr=1
      endif

c --- Derived data
      nsrc(k)=0
      do i=1,mxsrctype
         nsrc(k)=nsrc(k)+nsbytype(i,k)
      enddo
      ntotrec(k)=ndrec(k)+nctrec(k)
      if(lsamp(k)) then
         ngx(k)=meshdn(k)*(iesamp(k)-ibsamp(k))+1
         ngy(k)=meshdn(k)*(jesamp(k)-jbsamp(k))+1
         ntotrec(k)=ntotrec(k)+ngx(k)*ngy(k)
      else
         ngx(k)=0
         ngy(k)=0
      endif
      ngrec(k)=ngx(k)*ngy(k)
      ldisc(k)=.FALSE.
      if(ndrec(k).GT.0) ldisc(k)=.TRUE.
      lctsg(k)=.FALSE.
      if(nctrec(k).GT.0) lctsg(k)=.TRUE.

      return
      end

c-----------------------------------------------------------------------
      subroutine hd_proc(ioinp,k)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912           HD_PROC
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read remainder of a CALPUFF output file header
c
c --- UPDATES:
c
c --- v1.0 (120630) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0 (discrete receptor groups)
c
c --- INPUT:
c          IOINP - integer - File unit number for input CALPUFF file
c              k - integer - Sequential number for output CALPUFF file
c
c --- OUTPUT:
c
c --- HD_PROC called by: (main)
c --- HD_PROC calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header
      use userinf

      implicit none

c --- Arguments
      integer :: ioinp,k

c --- Locals
      character(len=16) :: units
      integer :: i,j,n,itype
      real    :: xsamLLkm, ysamLLkm, dsamkm

c --- Title (record NCOM+4)
      read(ioinp)  (title(i,k),i=1,3)

c --- Specie/level list (record NCOM+5)
      read(ioinp)  (csout(i,k),i=1,nspout(k))

c --- Units for each specie/level (record NCOM+5a)
      if(lspunit(k)) then
         read(ioinp)  (csunit(i,k),i=1,nspout(k))
      else
c ---    Set units based on dataset
         units='g/m3            '
         if(INDEX(dataset(k),'CONC.DAT').EQ.0)
     &       units='g/m2/s          '
         do i=1,nspout(k)
            csunit(i,k)=units
         enddo
      endif

c --- Non-gridded receptor coordinates (record NCOM+6)
c --- (New in dataset TNG-3.0: htrec, irgrp)
      if(ndrec(k).NE.0) then
         if(nrgrp(k).EQ.0) then
            read(ioinp) (xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k))
         elseif(nrgrp(k).GE.1) then
c ---       (TNG-3.0 and up record NCOM+6, 6a)
            read(ioinp) (xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k)),
     &                  (htrec(i,k),i=1,ndrec(k)),
     &                  (irgrp(i,k),i=1,ndrec(k))
            read(ioinp) (argrpnam(n,k),n=1,nrgrp(k))
         endif
      endif

c --- Complex terrain receptor coordinates (record NCOM+7)
      if(nctrec(k).NE.0) then
         read(ioinp)  (xctr(i,k),i=1,nctrec(k)),
     &                (yctr(i,k),i=1,nctrec(k)),
     &                (zctr(i,k),i=1,nctrec(k)),
     &                (ihill(i,k),i=1,nctrec(k))
      endif

c --- Source names (records NCOM+8 to NCOM+15)
c --- Loop over source types
      do i=1,mxsrctype
         if(nsbytype(i,k).GT.0) then
            read(ioinp)  itype,(csrcnam(n,i,k),n=1,nsbytype(i,k))
            if(itype.NE.i) then
c ---          Report problem information and quit
               write(*,*)
               write(*,*)
               write(*,*)'  *  Unexpected source type found, or   *'
               write(*,*)'  *  Unexpected source number found.    *'
               write(*,*)'  *  Source type expected: ',i
               write(*,*)'  *  Source type found   : ',itype
               write(*,*)'  *  Source Names Read:                 *'
               write(*,*) (csrcnam(n,i,k),n=1,nsbytype(i,k))
               stop
            endif
         endif
      enddo

c --- Derived data
c --- Coordinates of receptor vector
      n=0
c --- Compute coordinates of sampling grid receptors
      if(lsamp(k)) then
         xsamLLkm=xorigkm(k)+(FLOAT(ibsamp(k))-0.5)*dxkm(k)
         ysamLLkm=yorigkm(k)+(FLOAT(jbsamp(k))-0.5)*dykm(k)
         dsamkm=dxkm(k)/FLOAT(meshdn(k))
         do j=1,ngy(k)
         do i=1,ngx(k)
            n=n+1
            xreckm(n,k)=xsamLLkm+FLOAT(i-1)*dsamkm
            yreckm(n,k)=ysamLLkm+FLOAT(j-1)*dsamkm
         enddo
         enddo
      endif
c --- Add coordinates of discrete receptors
      if(ldisc(k)) then
         do i=1,ndrec(k)
            n=n+1
            xreckm(n,k)=xrec(i,k)
            yreckm(n,k)=yrec(i,k)
         enddo
      endif
c --- Add coordinates of CTSG receptors
      if(lctsg(k)) then
         do i=1,nctrec(k)
            n=n+1
            xreckm(n,k)=xctr(i,k)
            yreckm(n,k)=yctr(i,k)
         enddo
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine comp1(iolst,ioinp,ioda,ierr)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630             COMP1
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read all model data and store to DA file for processing
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input data file
c           IODA - integer - File unit number for DA storage file
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- COMP1 called by: (main)
c --- COMP1 calls:     GETPER, STOREDAY, TIME_DIFF, PUTDA, QADA
c----------------------------------------------------------------------
c --- Access module data
      use header
      use daccess
      use inpdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioinp,ioda,ierr

c --- Locals
      integer :: k,ndimx,ndimy,ndimd,ndimct
      integer :: ieof
      integer :: ksrc,delday,delhour,delminute,delsecond
      integer :: nsecpd, it, iseconds, ihh, iday, ndaycount

c --- Initializations
      ierr=0
      ieof=0
      ndaycount=0

c --- Array allocation should have resulted in exact match to
c --- shape actually used for all number of receptors --- verify this
      do k=1,1
         if(lsamp(k)) then
            ndimx=SIZE(datag,1)
            ndimy=SIZE(datag,2)
            if(ndimx.NE.ngx(k) .OR. ndimy.NE.ngy(k)) then
               write(iolst,*)'ERROR in COMP -- array allocation '//
     &                       'does not match gridded receptors in file'
               write(iolst,*)'   Allocated size: ',ndimx,ndimy
               write(iolst,*)'   Size in file  : ',ngx(k),ngy(k)
               ierr=1
               return
            endif
         endif
         if(ldisc(k)) then
            ndimd=SIZE(datad,1)
            if(ndimd.NE.ndrec(k)) then
               write(iolst,*)'ERROR in COMP -- array allocation '//
     &                       'does not match discrete receptors in file'
               write(iolst,*)'   Allocated size: ',ndimd
               write(iolst,*)'   Size in file  : ',ndrec(k)
               ierr=1
               return
            endif
         endif
         if(lctsg(k)) then
            ndimct=SIZE(datact,1)
            if(ndimct.NE.nctrec(k)) then
               write(iolst,*)'ERROR in COMP -- array allocation '//
     &                       'does not match CTSG receptors in file'
               write(iolst,*)'   Allocated size: ',ndimct
               write(iolst,*)'   Size in file  : ',nctrec(k)
               ierr=1
               return
            endif
         endif
      enddo

c --- Read all data from file k
      k=1

c --- Neglect any source contribution if present
c --- Source contribution configuration
      ksrc=1
      if(msource(k).GT.0) then
c ---    Skip individual sources, use total
         ksrc=nsrc(k)+1
      endif

c --- Number of seconds between periods read from data file
      nsecpd=iavg(k)*nsecdt(k)

c --- Initialize daily storage arrays
      val=0.0
      idatim=0
c --- Set times (HHSSSS) in each day since these are fixed
      idatim(1)=0
      do it=2,nvarrcrd-1
         iseconds=(it-1)*nsecpd
         ihh=iseconds/3600
         idatim(it)=ihh*10000+iseconds-ihh*3600
      enddo

c -----------------------------------------
c --- Process first period of data
c -----------------------------------------
      iday=1
c --- Get one period of data
      call GETPER(iolst,ioinp,k,ierr,ieof)
      if(ieof.NE.0) then
c ---    Bad file
         stop 'HALTED in COMP1: Bad data file (EOF on 1st read)'
      endif
      if(ierr.GT.0) stop 'HALTED in COMP1 -- See list file'
c --- Store into daily arrays
      call STOREDAY(iolst,k,ksrc,nsecpd,ierr)
      if(ierr.GT.0) stop 'HALTED in COMP1 -- See list file'

c --- Get averaging time from start-end times
      call TIME_DIFF(iolst,datetime_b(k),datetime_e(k),
     &               delday,delhour,delsecond,ierr)
      if(ierr.GT.0) stop 'HALTED in COMP1 -- See list file'
      delminute=delsecond/60
      delsecond=delsecond-delminute*60

c -----------------------------------------
c --- Process all time periods in the file
c -----------------------------------------
10    call GETPER(iolst,ioinp,k,ierr,ieof)
      if(ierr.GT.0) stop 'HALTED in COMP1 -- See list file'
      if(ieof.NE.0) then
c ---    Completed file
c ---    Do final output to DA file
         call PUTDA(iolst,ioda,iday,k)
         ndaycount=ndaycount+1
         goto 999
      endif

c --- Check for date change
      if((nyrb(k)*1000+njulb(k)).NE.idatim(nvarrcrd)) then
c ---    Store day-arrays in DA file
         call PUTDA(iolst,ioda,iday,k)
         ndaycount=ndaycount+1
c ---    Reset day-array for data
         val=0.0
c ---    Next day for DA file
         iday=iday+1
      endif

c --- Store current period into daily arrays
      call STOREDAY(iolst,k,ksrc,nsecpd,ierr)
      if(ierr.GT.0) stop 'HALTED in COMP1 -- See list file'

c --- Next period
      goto 10

999   continue

cc --- Check consistency
c      write(*,*) 'Number of days=ndaycount    :  ',ndaycount
c      write(*,*) 'Number of days in DA=numdays:  ',numdays

c --- Report selected information to list file
      write(iolst,*)
      write(iolst,*)'Information about data in file'
      write(iolst,*)
      write(iolst,*)'Averages (DAY,HH,MM,SS) =  ',delday,delhour,
     &                                            delminute,delsecond
      if(lrunning(k)) then
         write(iolst,*)'Type of Averaging       =  RUNNING'
      else
         write(iolst,*)'Type of Averaging       =  BLOCK'
      endif
      write(iolst,*)'Number of averages/day  = ',nvarrcrd-1

c --- Report information about DA file to list
      call QADA(iolst,ioda)

      return
      end

c-----------------------------------------------------------------------
      subroutine comp2(iolst,ioda,ierr)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 131113             COMP2
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Sort data in time by magnitude, for each species and
c               receptor, storing the values and date-time for each
c               requested rank.
c               Identify MAX of ranked values over all receptors.
c
c --- UPDATES:
c --- v1.1 (120630) to v1.2 (131113)
c           Change definition of the percentile to be consistent
c           with USEPA tabulations of percentiles and ranks
c           P(rank N of M values)=100.(1.-N/M)
c           N(percentile P of M values)=M(1.-P/100.) {round-up}
c             [e.g. N=7.01 rounds up to N=8]
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c           IODA - integer - File unit number for scratch DA file
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- COMP2 called by: (main)
c --- COMP2 calls:     SPSORT, DATESTRING, TIMESTRING
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use daccess
      use rankdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioda,ierr

c --- Locals
      logical :: ldaymax
      real    :: testpct
      integer :: i,j,k,is,ir,it,id
      integer :: it1,it2,is_step,idarec

c --- Initializations
      ierr=0
      k=1
      rankvalue=0.0
      arankdate='YYYY_JJJ'
      aranktime='HH:MM:SS'
      rankvaluemax=0.0
      arankdatemax='YYYY_JJJ'
      aranktimemax='HH:MM:SS'

c --- Set time range to rank each day
c --- ICDAY=0:  All times each day
c --- ICDAY=1:  Just MAX each day
      if(icday.EQ.0) then
         it1=1
         it2=nvarrcrd-1
         ldaymax=.FALSE.
      else
         it1=nvarrcrd
         it2=nvarrcrd
         ldaymax=.TRUE.
      endif

c --- Fill date vector for series (same for all species-receptors)
c --- Time vector is same every day as in last idatim read, but
c --- series will depend on rdatim if daily max is requested
      atserdate='YYYY_JJJ'
c --- Loop over days
      it=0
      do id=1,numdays
         idarec=(id-1)*nrcrdday+1
c ---    Times for 1 day in DA file, plus current date
         read(ioda,rec=idarec) idatim
c ---    Extract dates to working vector
         do i=it1,it2
            it=it+1
            call DATESTRING(idatim(nvarrcrd),atserdate(it))
         enddo
      enddo

c --- Set percentiles based on time periods used
c ----------------------------------------------
      ntimes=it
c --- Convert requested integer ranks to percentiles
      do i=1,n_nth

c --- v1.2(131113)
c        pct_from_nth(i)=100.*(1.0-(FLOAT(nth(i))-0.5)/FLOAT(ntimes))
         pct_from_nth(i)=100.*(1.0-FLOAT(nth(i))/FLOAT(ntimes))

      enddo
c --- Compute integer rank for requested percentiles
      do i=1,n_pct

c --- v1.2(131113)
c        nth_from_pct(i)=INT((1.0-0.01*pct(i))*FLOAT(ntimes)+0.5)
c        nth_from_pct(i)=MAX(nth_from_pct(i),1)
c        nth_from_pct(i)=MIN(nth_from_pct(i),ntimes)
c ---    Adjust if exact percentile exceeds requested value
c        if(nth_from_pct(i).LT.ntimes) then
c           testpct=100.*(1.0-(FLOAT(nth_from_pct(i))-0.5)/
c    &              FLOAT(ntimes))
c           if(testpct.GT.pct(i)) nth_from_pct(i)=nth_from_pct(i)+1
c        endif
         nth_from_pct(i)=INT((1.0-0.01*pct(i))*FLOAT(ntimes))
c ---    Adjust if exact percentile exceeds requested value
         if(nth_from_pct(i).LT.ntimes) then
            testpct=100.*(1.0-FLOAT(nth_from_pct(i))/FLOAT(ntimes))
            if(testpct.GT.pct(i)) nth_from_pct(i)=nth_from_pct(i)+1
         endif

      enddo

c --- Process all species and receptors in turn
c ---------------------------------------------
c --- Loop over species
      do is=1,nspout(k)
         write(*,*)'-- Ranking results for species-level: '//
     &              csout(is,k)
         is_step=(is-1)*ntotrec(k)
c ---    Loop over receptors
         do ir=1,ntotrec(k)

c ---       Reset working vectors
            tservalue=0.0
            atsertime='HH:MM:SS'
            irank=0

c ---       Extract values & times to working vectors
c ---       Loop over days
            it=0
            do id=1,numdays
               idarec=((id-1)*nrcrdday+1)+is_step+ir
c ---          Data for one day in DA file
               read(ioda,rec=idarec) rdatim
c ---          Loop over times on this day
               do i=it1,it2
                  it=it+1
                  tservalue(it)=rdatim(i)
                  if(ldaymax) then
c ---                Find time of maximum
                     call TIMESTRING(idatim(1),atsertime(it))
                     do j=1,nvarrcrd-1
                        if(rdatim(j).GE.rdatim(nvarrcrd))
     &                     call TIMESTRING(idatim(j),atsertime(it))
                     enddo
                  else
                     call TIMESTRING(idatim(i),atsertime(it))
                  endif
               enddo
            enddo

c           --------------------------------
c           --- Create ranked pointer vector
c           --- (SLATEC subroutine SPSORT)
c           --------------------------------

c ---       Sort from largest to smallest value
            i=-1
            call SPSORT(tservalue,it,irank,i,ierr)
            if(ierr.NE.0) then
               write(iolst,*)'ERROR in COMP2:  Sort failed with'//
     &                       ' ierr = ',ierr
               return
            endif

c ---       Store values-dates-times for requested nth-high ranks
            do i=1,n_nth
               rankvalue(ir,is,i)=tservalue(irank(nth(i)))
               arankdate(ir,is,i)=atserdate(irank(nth(i)))
               aranktime(ir,is,i)=atsertime(irank(nth(i)))
            enddo
c ---       Store values-dates-times for requested percentile ranks
            do i=1,n_pct
               rankvalue(ir,is,(i+n_nth))=
     &                           tservalue(irank(nth_from_pct(i)))
               arankdate(ir,is,(i+n_nth))=
     &                           atserdate(irank(nth_from_pct(i)))
               aranktime(ir,is,(i+n_nth))=
     &                           atsertime(irank(nth_from_pct(i)))
            enddo
         enddo
      enddo

c ------------------------------------------------------------
c --- Find the max for each requested rank, over all receptors
c --- for each species (store corresponding date, time)
c ------------------------------------------------------------

      rankvaluemax=0.0
c --- Loop over requested ranks
      do i=1,n_ranks
c ---    Loop over species
         do is=1,nspout(k)
c ---       Loop over receptors
            do ir=1,ntotrec(k)
               if(rankvalue(ir,is,i).GE.rankvaluemax(is,i)) then
                  rankvaluemax(is,i)=rankvalue(ir,is,i)
                  arankdatemax(is,i)=arankdate(ir,is,i)
                  aranktimemax(is,i)=aranktime(ir,is,i)
                  rankxmax(is,i)=xreckm(ir,k)
                  rankymax(is,i)=yreckm(ir,k)
               endif
            enddo
         enddo
      enddo

      return
      end

c-----------------------------------------------------------------------
      subroutine storeday(iolst,k,ksrc,nsecpd,ierr)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630          STOREDAY
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Store data read for 1 period (source ksrc)
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c              K - integer - Current data file
c           KSRC - integer - Current source
c         NSECPD - integer - Number of seconds between time periods
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- STOREDAY called by: COMP1
c --- STOREDAY calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use daccess
      use header
      use inpdata

      implicit none

c --- Declare arguments
      integer :: iolst,k,ksrc,nsecpd,ierr

c --- Locals
      integer :: it, is, ir, i, j

c --- Date goes in last index of date-time array
      idatim(nvarrcrd)=nyrb(k)*1000+njulb(k)

c --- Time index
      it=(nhrb(k)*3600+nsecb(k))/nsecpd +1
      if(it.LE.0 .OR. it.GE.nvarrcrd) then
         write(iolst,*)'ERROR in STOREDAY -- time period index '//
     &                 'is out of expected range'
         write(iolst,*)'   Expected: 1 to ',nvarrcrd-1
         write(iolst,*)'   Found   : ',it
         ierr=1
         return
      endif

c --- Store data for this period (update the day-max also)
c --- Loop on species
      do is=1,nspout(k)
         ir=0
c ---    Gridded receptors
         if(lsamp(k)) then
            do j=1,ngy(k)
            do i=1,ngx(k)
               ir=ir+1
               val(it,ir,is)=datag(i,j,is,ksrc,k)
               val(nvarrcrd,ir,is)=MAX(val(nvarrcrd,ir,is),
     &                                 val(it,ir,is))
            enddo
            enddo
         endif
c ---    Discrete receptors
         if(ldisc(k)) then
            do i=1,ndrec(k)
               ir=ir+1
               val(it,ir,is)=datad(i,is,ksrc,k)
               val(nvarrcrd,ir,is)=MAX(val(nvarrcrd,ir,is),
     &                                 val(it,ir,is))
            enddo
         endif
c ---    CTSG receptors
         if(lctsg(k)) then
            do i=1,nctrec(k)
               ir=ir+1
               val(it,ir,is)=datact(i,is,ksrc,k)
               val(nvarrcrd,ir,is)=MAX(val(nvarrcrd,ir,is),
     &                                 val(it,ir,is))
            enddo
         endif
      enddo 

      return
      end

c-----------------------------------------------------------------------
      subroutine putda(iolst,ioda,iday,k)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630             PUTDA
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Write data for 1 day to DA file
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c           IODA - integer - File unit number for DA storage file
c           IDAY - integer - Sequential day-number for DA file
c              K - integer - Current data file
c
c --- OUTPUT:
c
c --- PUTDA called by: COMP1
c --- PUTDA calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use daccess
      use header
      use rankdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioda,iday,k

c --- Locals
      integer :: nr, is, ir, it

c --- Set first record index
      nr=(iday-1)*nrcrdday+1

c --- Write the date-time record
      write(ioda,rec=nr)  idatim

c --- Loop over species
      do is=1,nspout(k)
c ---    Loop over receptors
         do ir=1,ntotrec(k)
            nr=nr+1
            write(ioda,rec=nr) (val(it,ir,is),it=1,nvarrcrd)
         enddo
      enddo

c --- Tallies
c --- Number of days stored in DA file
      numdays=iday
c --- Number of times stored in DA file
      ntimesda=(nvarrcrd-1)*numdays
c --- Number of records in DA file
      nrcrdfile=nr

      return
      end

c-----------------------------------------------------------------------
      subroutine qada(iolst,ioda)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630              QADA
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Check/report data stored in DA file
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c           IODA - integer - File unit number for DA storage file
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- QADA called by: COMP1
c --- QADA calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use daccess
      use header

      implicit none

c --- Declare arguments
      integer :: iolst,ioda

c --- Locals
      integer :: i,irec

      write(iolst,*)'Number of days          = ',numdays
      write(iolst,*)'Number of receptors     = ',ntotrec(1)
      write(iolst,*)'Number of species       = ',nspout(1)
c      write(iolst,*)'Number of records in file = ',nrcrdfile
c      write(iolst,*)'Number of records per day = ',nrcrdday

c --- Grab first record
      read(ioda,rec=1) idatim
      write(iolst,*)'First day (YYYYJJJ)     = ',idatim(nvarrcrd),
     &              '    '//axbtz(1)
c      write(iolst,*)'Times (HHSSSS)          = ',(idatim(i),
c     &               i=1,nvarrcrd-1)

c --- Grab last date-time record
      irec=nrcrdfile-nrcrdday+1
      read(ioda,rec=irec) idatim
      write(iolst,*)'Last day (YYYYJJJ)      = ',idatim(nvarrcrd),
     &              '    '//axbtz(1)
c      write(iolst,*)'Times (HHSSSS)          = ',(idatim(i),
c     &               i=1,nvarrcrd-1)


      write(iolst,*)
      write(iolst,*)

      return
      end

c-----------------------------------------------------------------------
      subroutine getper(iolst,ioinp,k,ierr,ieof)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630            GETPER
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read and store model data records for 1 time period
c
c               Receptor arrays must be dimensioned EXACTLY for the
c               number(s) of receptors in the CALPUFF file!
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input CALPUFF file
c              K - integer - Current CALPUFF file index
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c           IEOF - integer - End of file flag
c
c --- GETPER called by: COMP1
c --- GETPER calls:     DATE_HRSEC, CMPRSIO, RDBDAT
c----------------------------------------------------------------------
c --- Access module data
      use header
      use inpdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioinp,k,ierr,ieof

c --- Locals
      integer :: is,nsources,isp
      integer :: ncmprs,ione

c --- Initializations
      ione=0
      ieof=0
      ierr=0

c --- Source contribution configuration
      if(msource(k).EQ.0) then
c ---    Sum of all sources
         nsources=1
      else
c ---    Individual sources plus total
         nsources=nsrc(k)+1
      endif

c --- Source loop
c ---------------
      do is=1,nsources

c ---    Record 1:  Time period
         read(ioinp,end=999)
     &        nyrb(k),njulb(k),nhrb(k),nsecb(k),
     &        nyre(k),njule(k),nhre(k),nsece(k)
         call DATE_HRSEC(iolst,'TOSNGL',
     &                   nyrb(k),njulb(k),nhrb(k),nsecb(k),
     &                   datetime_b(k),ierr)
         call DATE_HRSEC(iolst,'TOSNGL',
     &                   nyre(k),njule(k),nhre(k),nsece(k),
     &                   datetime_e(k),ierr)
         if(ierr.NE.0) stop 'HALTED in GETPER -- See list file'

c ---    Record 2:  Source information
         read(ioinp,end=999)
     &        istype(is,k),isnum(is,k),sname(is,k),
     &        sxkm(is,k),sykm(is,k)

c ---    Output species-level loop
c --------------------------------
         do isp=1,nspout(k)
            if(lsamp(k)) then
               if(lcompr(k)) then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datag(1,1,isp,is,k),ngrec(k),
     &                         workg,ncmprs,cspecg(isp,k),ierr)
                  if(ierr.NE.0) stop 'HALTED in GETPER -- See list file'
               else
                  call RDBDAT(ioinp,cspecg(isp,k),
     &                        datag(1,1,isp,is,k),ngx(k),ngy(k))
               endif
            endif

            if(ldisc(k)) then
               if(lcompr(k))then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datad(1,isp,is,k),ndrec(k),
     &                         workd,ncmprs,cspecd(isp,k),ierr)
                  if(ierr.NE.0) stop 'HALTED in GETPER -- See list file'
               else
                  call RDBDAT(ioinp,cspecd(isp,k),
     &                        datad(1,isp,is,k),ndrec(k),ione)
               endif
            endif

            if(lctsg(k)) then
               if(lcompr(k))then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datact(1,isp,is,k),nctrec(k),
     &                         workct,ncmprs,cspecct(isp,k),ierr)
                  if(ierr.NE.0) stop 'HALTED in GETPER -- See list file'
               else
                  call RDBDAT(ioinp,cspecct(isp,k),
     &                        datact(1,isp,is,k),nctrec(k),ione)
               endif
            endif
         enddo
c --------------------------------
c ---    End output species-level loop

      enddo
c -------------------
c --- End source loop

      return

999   ieof=1
      write(*,*)'GETPER:  Finished reading data file '
      
      return

      end

c-----------------------------------------------------------------------
      subroutine output(io,ioplt)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120915            OUTPUT
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Report MAX results to list file and create a plot file
c               for each requested rank/percentile that lists the
c               concentration of each species at each receptor
c
c --- UPDATES:
c --- v1.0 (120630) to v1.1 (120915)
c        1. Change file pathname variables from C*132 to C*256
c        2. Increase number of digits in percentile filename
c        3. Format for output percentile values changed from 3x,f5.1,5x
c           to 2x,f7.3,4x format
c        4. Add user-selected mass units for output
c
c --- INPUT:
c             IO - integer - File unit number for output List file
c          IOPLT - integer - File unit number for output Plot file(s)
c
c --- OUTPUT:
c
c --- OUTPUT called by: (main)
c --- OUTPUT calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use rankdata

      implicit none

c --- Declare arguments
      integer :: io,ioplt

c --- Locals
      integer :: i,j,k,is
      character(len=256) :: plotfile
      character(len=8)   :: datestring
      character(len=16)  :: cunits(5), dunits(5)
      real               :: xaunits(5)

c --- Define mass units label
      data cunits/'g/m3            ','mg/m3           ',
     &            'ug/m3           ','ng/m3           ',
     &            'pg/m3           '/
      data dunits/'g/m2/s          ','mg/m2/s         ',
     &            'ug/m2/s         ','ng/m2/s         ',
     &            'pg/m2/s         '/

c --- Define mass units multiplier
      data xaunits/1.0, 1.0e03, 1.0e06, 1.0e09, 1.0e12/
      
c --- Initializations
      k=1

c --- Set up units label and multiplier (user request) arrays
      counit=csunit
      xounit=1.0
c --- Units are changed only if they are the standard g/* and
c --- if user selected an alternate scaling
      if(mass_u.GT.1) then
         do is=1,nspout(k)
            if(csunit(is,k).EQ.cunits(1)) then
               counit(is,k)=cunits(mass_u) 
               xounit(is,k)=xaunits(mass_u) 
            elseif(csunit(is,k).EQ.dunits(1)) then
               counit(is,k)=dunits(mass_u) 
               xounit(is,k)=xaunits(mass_u) 
            endif
         enddo
      endif

c --- Info to list file
      if(icday.EQ.0) then
         write(io,*)'Model data for each receptor are ranked over'//
     &              ' all times based on magnitude and'
         write(io,*)'the requested Nth-Highest and/or Percentile '//
     &              'values are reported for each receptor'
         write(io,*)'in the output files:'
      else
         write(io,*)'Calendar-day maximum for each receptor are'//
     &              ' ranked over all days based on magnitude and'
         write(io,*)'the requested Nth-Highest and/or Percentile '//
     &              'values are reported for each receptor'
         write(io,*)'in the output files:'
      endif
      write(io,*)

c --- Output Plot files for requested ranks of maximum at all receptors
      j=LEN_TRIM(lstfile)
      do i=1,n_nth
c ---    Create filename
         plotfile=' '
         plotfile=TRIM(lstfile)//'_PLOT_RANK-nnnn.DAT'
         write(plotfile((j+12):(j+15)),'(i4.4)') nth(i)
         open(ioplt,file=plotfile,status='unknown')
         call XYCOUT(ioplt,i,nspout(k))
c ---    Report name to list file
         write(io,*)TRIM(plotfile)
         close(ioplt)
      enddo
      do i=1,n_pct
c ---    Create filename
         plotfile=' '
         if(pct(i).LE.9.999) then
            plotfile=TRIM(lstfile)//'_PLOT_PCTL-p.ppp.DAT'
            write(plotfile((j+12):(j+16)),'(f5.3)') pct(i)
         elseif(pct(i).LE.99.999) then
            plotfile=TRIM(lstfile)//'_PLOT_PCTL-pp.ppp.DAT'
            write(plotfile((j+12):(j+17)),'(f6.3)') pct(i)
         else
            plotfile=TRIM(lstfile)//'_PLOT_PCTL-100.0.DAT'
         endif
         open(ioplt,file=plotfile,status='unknown')
         call XYCOUT(ioplt,(i+n_nth),nspout(k))
c ---    Report name to list file
         write(io,*)TRIM(plotfile)
         close(ioplt)
      enddo

c --- Ouput table for requested ranks of maximum over all receptors
      write(io,*)
      write(io,*)
      write(io,*)'The largest of these for each rank are tabulated '//
     &           'below with the location and time'


c --- Nth-Highest
      if(n_nth.GT.0) then
      write(io,*)
      write(io,*)
      write(io,'(15x,a44)')
     &   '--------------------------------------------'
      write(io,'(15x,a44)')
     &   'Tabulation for Requested Nth-Highest Values '
      write(io,'(15x,a44)')
     &   'Results for the Maximum Receptor            '
      if(icday.EQ.0) then
         write(io,'(15x,a44)')
     &      'All Values in Each Day are Included         '
      else
         write(io,'(15x,a44)')
     &      'Only daily maximums at receptor are Included'
      endif
      write(io,'(15x,a44)')
     &   '--------------------------------------------'
      write(io,'(a106)')
     &   '                                                         '//
     &   '                 Location       Starting '//axbtz(k)
      write(io,'(a108)')
     &   '                                                         '//
     &   '                X         Y       Date      Time   '
      write(io,'(a108)')
     &   '    N Percentile  Species - Level     Modeled      Units '//
     &   '              (KM)      (KM)   (YYYY_JJJ)(HH:MM:SS)'
      write(io,*)
      do is=1,nspout(k)
        do i=1,n_nth
c         write(io,'(i5,3x,f5.1,5x,a15,1pe16.7,2x,a16,0p2f10.3,2i10.5)')
c     &           nth(i),pct_from_nth(i),csout(is,k),
c     &           xounit(is,k)*rankvaluemax(is,i),
c     &           counit(is,k),rankxmax(is,i),rankymax(is,i),
c     &           irankdatemax(is,i),iranktimemax(is,i)
         write(io,'(i5,2x,f7.3,4x,a15,1pe16.7,2x,a16,0p2f10.3,2a10)')
     &           nth(i),pct_from_nth(i),csout(is,k),
     &           xounit(is,k)*rankvaluemax(is,i),
     &           counit(is,k),rankxmax(is,i),rankymax(is,i),
     &           arankdatemax(is,i),aranktimemax(is,i)
        enddo
      enddo
      write(io,*)
      write(io,'(a44)')
     &   '--------------------------------------------'
      write(io,'(a34,i6)')
     &   'Note:  number of times in period = ',ntimes
      write(io,*)
      write(io,*)
      endif

c --- Percentile
      if(n_pct.GT.0) then
      write(io,*)
      write(io,*)
      write(io,'(15x,a44)')
     &   '--------------------------------------------'
      write(io,'(15x,a44)')
     &   'Tabulation for Requested Percentile Values  '
      write(io,'(15x,a44)')
     &   'Results for the Maximum Receptor            '
      if(icday.EQ.0) then
         write(io,'(15x,a44)')
     &      'All Values in Each Day are Included         '
      else
         write(io,'(15x,a44)')
     &      'Only daily maximums at receptor are Included'
      endif
      write(io,'(15x,a44)')
     &   '--------------------------------------------'
      write(io,'(a106)')
     &   '                                                         '//
     &   '                 Location       Starting '//axbtz(k)
      write(io,'(a108)')
     &   '                                                         '//
     &   '                X         Y       Date      Time   '
      write(io,'(a108)')
     &   '    N Percentile  Species - Level     Modeled      Units '//
     &   '              (KM)      (KM)   (YYYY_JJJ)(HH:MM:SS)'
      write(io,*)
      do is=1,nspout(k)
        do i=1,n_pct
         j=i+n_nth
c         write(io,'(i5,3x,f5.1,5x,a15,1pe16.7,2x,a16,0p2f10.3,2i10.5)')
c     &           nth_from_pct(i),pct(i),csout(is,k),
c     &           xounit(is,k)*rankvaluemax(is,j),
c     &           counit(is,k),rankxmax(is,j),rankymax(is,j),
c     &           irankdatemax(is,j)),iranktimemax(is,j)
         write(io,'(i5,2x,f7.3,4x,a15,1pe16.7,2x,a16,0p2f10.3,2a10)')
     &           nth_from_pct(i),pct(i),csout(is,k),
     &           xounit(is,k)*rankvaluemax(is,j),
     &           counit(is,k),rankxmax(is,j),rankymax(is,j),
     &           arankdatemax(is,j),aranktimemax(is,j)
        enddo
      enddo
      write(io,*)
      write(io,'(a44)')
     &   '--------------------------------------------'
      write(io,'(a34,i6)')
     &   'Note:  number of times in period = ',ntimes
      write(io,*)
      write(io,*)
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine xycout(io,nrank,nsp)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120915            XYCOUT
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Write MAX results for each receptor to plot file
c               for each requested rank/percentile that lists the
c               concentration of each species at each receptor
c
c --- UPDATES:
c --- v1.0 (120630) to v1.1 (120915)
c        1. Percentile written to 3 decimal places
c        2. Add user-selected mass units for output
c
c --- INPUT:
c             IO - integer - File unit number for output Plot file
c          NRANK - integer - Rank to output (1 to n_ranks)
c            NSP - integer - Number of species-levels (for local array)
c
c --- OUTPUT:
c
c --- XYCOUT called by: (main)
c --- XYCOUT calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use rankdata

      implicit none

c --- Declare arguments
      integer :: io,nrank,nsp

c --- Locals
      integer :: k,ir,is,i
      character(len=20)   :: head1,head2,head3
      character(len=40)   :: add1x,add2x,add3x
      character(len=40)   :: add1(nsp),add2(nsp)
      character(len=40)   :: add3(nsp)
      
c --- Initializations
      k=1

c --- Construct column headers
      head1(1:20)='       Location     '
      head2(1:20)='      X         Y   '
      head3(1:20)='    (KM)      (KM)  '
      add1x='    Species - Level  Starting UTC-0700  '
      add2x='    SO2           1    Date     Time    '
      add3x='    g/m3            (YYYY_JJJ)(HH:MM:SS)'
c --- Drop in the UTC zone identifier
      add1x(31:38)=axbtz(k)
c --- Make segment for each species
      do is=1,nspout(k)
c ---    Start with the template
         add1(is)=add1x
         add2(is)=add2x
         add3(is)=add3x
c ---    Add species information
         add2(is)(5:19)=csout(is,k)
         add3(is)(5:19)=counit(is,k)
      enddo

c --- Banner
c ----------
      if(n_nth.GT.0 .AND. nrank.LE.n_nth) then
c ---    Nth-Highest
         i=nrank
         write(io,*)
         write(io,*)
         write(io,'(15x,a45)')
     &      '---------------------------------------------'
         write(io,'(15x,a38,i7)')
     &      'Requested Nth-Highest Values for N  = ',nth(i)
         write(io,'(15x,a38,f7.3)')
     &      'Corresponding Percentile (%)        = ',pct_from_nth(i)
         if(icday.EQ.0) then
            write(io,'(15x,a45)')
     &         'All Values in Each Day are Included          '
         else
            write(io,'(15x,a45)')
     &         'Only daily maximums at receptor are Included '
         endif
         write(io,'(15x,a38,i7)')
     &   'Number of times in period           = ',ntimes
         write(io,'(15x,a45)')
     &      '---------------------------------------------'
         write(io,*)
         write(io,*)
      else
c ---    Percentile
         i=nrank-n_nth
         write(io,*)
         write(io,*)
         write(io,'(15x,a45)')
     &      '---------------------------------------------'
         write(io,'(15x,a38,f7.3)')
     &      'Requested Values for Percentile (%) = ',pct(i)
         write(io,'(15x,a38,i7)')
     &      'Corresponding Nth-Highest N         = ',nth_from_pct(i)
         if(icday.EQ.0) then
            write(io,'(15x,a45)')
     &         'All Values in Each Day are Included          '
         else
            write(io,'(15x,a45)')
     &         'Only daily maximums at receptor are Included '
         endif
         write(io,'(15x,a38,i7)')
     &   'Number of times in period           = ',ntimes
         write(io,'(15x,a45)')
     &      '---------------------------------------------'
         write(io,*)
         write(io,*)
      endif

c --- More Info to file header
      if(icday.EQ.0) then
         write(io,*)'Model data for each receptor are ranked over'//
     &              ' all times based on magnitude and'
         write(io,*)'the requested Nth-Highest and/or Percentile '//
     &              'values are reported for each receptor'
      else
         write(io,*)'Calendar-day maxima for each receptor are'//
     &              ' ranked over all days based on magnitude and'
         write(io,*)'the requested Nth-Highest and/or Percentile '//
     &              'values are reported for each receptor'
      endif
      write(io,*)
      write(io,*)

c --- Column headings
c -------------------
      write(io,*)
      write(io,'(a20,1000a40)') head1,(add1(is),is=1,nspout(k))
      write(io,'(a20,1000a40)') head2,(add2(is),is=1,nspout(k))
      write(io,'(a20,1000a40)') head3,(add3(is),is=1,nspout(k))
      write(io,*)

c --- Loop over receptors to write data records
      do ir=1,ntotrec(k)
         write(io,'(2f10.3,1p1000(2x,e16.7,1x,2a10,1x))')
     &                   xreckm(ir,k),yreckm(ir,k),
     &                  (xounit(is,k)*rankvalue(ir,is,nrank),
     &                   arankdate(ir,is,nrank),
     &                   aranktime(ir,is,nrank),is=1,nspout(k))
      enddo

      return
      end


c-----------------------------------------------------------------------
      subroutine datestring(iyyyyjjj,ayyyyjjj)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630        DATESTRING
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Change date from integer to string
c
c --- INPUT:
c       IYYYYJJJ - integer - Year, Julian Day as integer
c
c --- OUTPUT:
c       AYYYYJJJ - char*8  - YYYY_JJJ string
c
c --- DATESTRING called by: OUTPUT
c --- DATESTRING calls:     none
c----------------------------------------------------------------------

      implicit none

      integer           :: i,j, iyyyyjjj
      character(len=8)  :: ayyyyjjj


c       write(*,*) 'DATESTRING  input: ',iyyyyjjj


      write(ayyyyjjj(1:7),'(i7.7)') iyyyyjjj
      do i=7,5,-1
         j=i+1
         ayyyyjjj(j:j)=ayyyyjjj(i:i)
      enddo
      ayyyyjjj(5:5)='_'


c       write(*,*) 'DATESTRING output: ',ayyyyjjj

      return
      end

c-----------------------------------------------------------------------
      subroutine timestring(itime,atime)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630        TIMESTRING
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Change time from integer to string
c
c --- INPUT:
c          ITIME - integer - HHSSSS
c
c --- OUTPUT:
c          ATIME - char*8  - HH:MM:SS string
c
c --- TIMETRING called by: OUTPUT
c --- TIMESTRING calls:     none
c----------------------------------------------------------------------

      implicit none

      integer           :: ih,im,is, itime
      character(len=8)  :: atime

c       write(*,*) 'TIMESTRING  input: ',itime

      ih=itime/10000
      is=itime-ih*10000
      im=is/60
      is= is-im*60
      atime='HH:MM:SS'
      write(atime(1:2),'(i2.2)') ih
      write(atime(4:5),'(i2.2)') im
      write(atime(7:8),'(i2.2)') is

c       write(*,*) 'TIMESTRING output: ',atime

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_1(k)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912           ALLOC_1
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays of header information that depend only
c               on the number of CALPUFF files processed
c               Arrays for user control inputs are allocated in CONTROL
c
c --- UPDATES:
c
c --- v1.0 (120630) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c              k - integer - Number of CALPUFF files
c
c --- OUTPUT:
c
c --- ALLOC_1 called by: (main)
c --- ALLOC_1 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header

      implicit none

c --- Arguments
      integer :: k

c --- Logical arrays (read from file)
      allocate (lcompr(k),lsamp(k))

c --- Logical arrays (derived)
      allocate (ldisc(k),lctsg(k),lspunit(k),lrunning(k))

c --- Character arrays (read from file)
      allocate (cmodel(k),ver(k),level(k))
      allocate (dataset(k),dataver(k),datamod(k))
      allocate (title(3,k))
      allocate (pmap(k),datum(k),axbtz(k),daten(k),utmhem(k))
      allocate (clat0(k),clon0(k),clat1(k),clat2(k))

c --- Integer arrays (read from file)
      allocate (ncom(k),ibyr(k),ibjul(k),ibhr(k),ibsec(k),irlg(k))
      allocate (iavg(k),nsecdt(k),nxm(k),nym(k),nz(k),nssta(k))
      allocate (ibcomp(k),iecomp(k),jbcomp(k),jecomp(k))
      allocate (ibsamp(k),jbsamp(k),iesamp(k),jesamp(k))
      allocate (meshdn(k),nsrctype(k),msource(k))
      allocate (ndrec(k),nrgrp(k),nctrec(k),nspout(k))
      allocate (i2dmet(k),iutmzn(k))
      allocate (nsbytype(mxsrctype,k))

c --- Integer arrays (derived)
      allocate (nsrc(k),ngx(k),ngy(k),ngrec(k),ntotrec(k))

c --- Real arrays (read from file)
      allocate (xbtz(k),dxkm(k),dykm(k),xorigkm(k),yorigkm(k))
      allocate (feast(k),fnorth(k))
      allocate (rnlat0(k),relon0(k),xlat1(k),xlat2(k))

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_2(k)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912           ALLOC_2
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays of header information that depend on
c               the number of CALPUFF files processed and other
c               configuration data, and all data arrays
c
c --- UPDATES:
c
c --- v1.1 (120915) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- v1.0 (120630) to v1.1 (120915)
c        1. Add user output units arrays COUNIT(), XOUNIT()
c
c --- INPUT:
c              k - integer - Number of CALPUFF files
c
c --- OUTPUT:
c
c --- ALLOC_2 called by: (main)
c --- ALLOC_2 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use inpdata
      use daccess

      implicit none

c --- Arguments
      integer :: k

c --- Locals
      integer :: i,n
      integer :: mxspout,mxsrc,mxsrc1,mxsrc2
      integer :: mxdrec,mxctrec,mxnx,mxny,mxrec,mxrgrp

c --------------------------
c --- Set MAX's across files
c --------------------------

c --- Get maximum number of species-levels across files
      mxspout=1
      do i=1,k
         mxspout=MAX(mxspout,nspout(i))
      enddo

c --- Get maximum number of discrete receptors across files
      mxdrec=1
      do i=1,k
         mxdrec=MAX(mxdrec,ndrec(i))
      enddo

c --- Get maximum number of gridded receptors across files
      mxnx=1
      mxny=1
      do i=1,k
         mxnx=MAX(mxnx,ngx(i))
         mxny=MAX(mxny,ngy(i))
      enddo

c --- Get maximum number of CTSG receptors across files
      mxctrec=1
      do i=1,k
         mxctrec=MAX(mxctrec,nctrec(i))
      enddo

c --- Get maximum number of all receptors across files
      mxrec=1
      do i=1,k
         mxrec=MAX(mxrec,(ndrec(i)+nctrec(i)+ngx(i)*ngy(i)))
      enddo

c --- Get maximum number of each source type across files
      mxsrc=1
      mxsrc1=1
      mxsrc2=1
      do i=1,k
         mxsrc=MAX(mxsrc,nsrc(i))
         do n=1,mxsrctype
            mxsrc1=MAX(mxsrc1,nsbytype(n,i))
         enddo
         if(msource(i).EQ.1) mxsrc2=MAX(mxsrc2,nsrc(i)+1)
      enddo

c --- Get maximum number of receptor groups across files
      mxrgrp=1
      do i=1,k
         mxrgrp=MAX(mxrgrp,nrgrp(i))
      enddo

c -------------------
c --- Set array sizes
c -------------------

c --- Character arrays for header
      allocate (csout(mxspout,k),csunit(mxspout,k))
      allocate (csrcnam(mxsrc1,mxsrctype,k))
      allocate (argrpnam(mxrgrp,k))

c --- Character arrays for header (derived)
c --- Modified units names
      allocate (counit(mxspout,k))

c --- Integer arrays for header
      allocate (ihill(mxctrec,k))
      allocate (irgrp(mxdrec,k))
      
c --- Real arrays for header
      allocate (xrec(mxdrec,k),yrec(mxdrec,k))
      allocate (zrec(mxdrec,k),htrec(mxdrec,k))
      allocate (xctr(mxctrec,k),yctr(mxctrec,k),zctr(mxctrec,k))

c --- Real arrays for header (derived)
      allocate (xreckm(mxrec,k),yreckm(mxrec,k))
c --- Modified units multiplier
      allocate (xounit(mxspout,k))

c --- Character arrays for data
      allocate (sname(mxsrc,k))
      allocate (cspecg(mxspout,k),cspecd(mxspout,k),cspecct(mxspout,k))

c --- Integer arrays for data
      allocate (istype(mxsrc,k),isnum(mxsrc,k))
      allocate (nyrb(k),njulb(k))
      allocate (nhrb(k),nsecb(k))
      allocate (nyre(k),njule(k))
      allocate (nhre(k),nsece(k))
      allocate (datetime_b(k),datetime_e(k))

c --- Real arrays for data
      allocate (sxkm(mxsrc,k),sykm(mxsrc,k))
      allocate (datag(mxnx,mxny,mxspout,mxsrc2,k))
      allocate (datad(mxdrec,mxspout,mxsrc2,k))
      allocate (datact(mxctrec,mxspout,mxsrc2,k))

c --- Real work arrays
      allocate (workg(mxnx,mxny))
      allocate (workd(mxdrec))
      allocate (workct(mxctrec))

c --- Integer arrays for daccess
      allocate (idatim(nvarrcrd))
c --- Real arrays for daccess
      allocate (rdatim(nvarrcrd))
      allocate (val(nvarrcrd,ntotrec(k),nspout(k)))

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_3(k)
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630           ALLOC_3
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays for assembling requested ranks
c
c --- INPUT:
c              k - integer - Number of the CALPUFF array to use
c
c --- OUTPUT:
c
c --- ALLOC_3 called by: (main)
c --- ALLOC_3 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use rankdata

      implicit none
      integer :: k

c --- Working vectors for 1 species, 1 receptor
      allocate (irank(ntimesda))
      allocate (atserdate(ntimesda),atsertime(ntimesda))
      allocate (tservalue(ntimesda))

c --- Array of values (all species, receptors) for requested ranks
      allocate (rankvalue(ntotrec(k),nspout(k),n_ranks))
      allocate (arankdate(ntotrec(k),nspout(k),n_ranks))
      allocate (aranktime(ntotrec(k),nspout(k),n_ranks))

c --- Array of MAX over receptors (all species) for requested ranks
      allocate (rankvaluemax(nspout(k),n_ranks))
      allocate (rankxmax(nspout(k),n_ranks))
      allocate (rankymax(nspout(k),n_ranks))
      allocate (arankdatemax(nspout(k),n_ranks))
      allocate (aranktimemax(nspout(k),n_ranks))

      return
      end

c-----------------------------------------------------------------------
      subroutine dealloc_1
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 120630         DEALLOC_1
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  De-Allocate arrays for model data input after DA file
c               is created
c
c --- INPUT:
c
c --- OUTPUT:
c
c --- DEALLOC_1 called by: (main)
c --- DEALLOC_1 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use inpdata
      use daccess

      implicit none

c --- Real arrays for inpdata
      deallocate (sxkm,sykm)
      deallocate (datag,datad,datact)

c --- Real work arrays for inpdata
      deallocate (workg,workd,workct)

c --- Real work arrays for daccess
      deallocate (val)

      return
      end

c-----------------------------------------------------------------------
      subroutine dealloc_2
c-----------------------------------------------------------------------
c
c --- CALRANK   Version: TNG-2.0.0      Level: 140912         DEALLOC_2
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  De-Allocate arrays
c
c --- UPDATES:
c
c --- v1.1 (120915) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- v1.0 (120630) to v1.1 (120915)
c        1. Add user output units arrays COUNIT(), XOUNIT()
c
c --- INPUT:
c
c --- OUTPUT:
c
c --- DEALLOC_2 called by: (main)
c --- DEALLOC_2 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use inpdata
      use rankdata
      use daccess

      implicit none

c --- Logical arrays for header
      deallocate (lcompr,lsamp,ldisc,lctsg,lspunit,lrunning)

c --- Character arrays for header
      deallocate (cmodel,ver,level)
      deallocate (dataset,dataver,datamod)
      deallocate (pmap,datum,axbtz,daten,utmhem)
      deallocate (clat0,clon0,clat1,clat2)
      deallocate (title,csout,csunit)
      deallocate (csrcnam,argrpnam)
c --- Modified units names
      deallocate (counit)

c --- Integer arrays for header
      deallocate (ncom,ibyr,ibjul,ibhr,ibsec,irlg)
      deallocate (iavg,nsecdt,nxm,nym,nz,nssta)
      deallocate (ibcomp,iecomp,jbcomp,jecomp)
      deallocate (ibsamp,jbsamp,iesamp,jesamp)
      deallocate (meshdn,msource)
      deallocate (nsrctype,nsbytype,nrgrp,irgrp)
      deallocate (ndrec,nctrec,nspout,i2dmet,iutmzn)
      deallocate (ihill)
      deallocate (nsrc,ngx,ngy,ntotrec)
      
c --- Real arrays for header
      deallocate (xbtz,dxkm,dykm,xorigkm,yorigkm)
      deallocate (feast,fnorth)
      deallocate (rnlat0,relon0,xlat1,xlat2)
      deallocate (xrec,yrec,zrec,htrec)
      deallocate (xctr,yctr,zctr)
      deallocate (xreckm,yreckm)
c --- Modified units multiplier
      deallocate (xounit)


c --- Character arrays for inpdata
      deallocate (sname,cspecg,cspecd,cspecct)

c --- Integer arrays for inpdata
      deallocate (istype,isnum)
      deallocate (nyrb,njulb,nhrb,nsecb)
      deallocate (nyre,njule,nhre,nsece)

c --- Integer arrays for userinf
      deallocate (nth,nth_from_pct)
c --- Real arrays for userinf
      deallocate (pct,pct_from_nth)

c --- Integer arrays for rankdata
      deallocate (irank)
c --- Character arrays for rankdata
      deallocate (atserdate,atsertime,arankdate,aranktime)
      deallocate (arankdatemax,aranktimemax)

c --- Real arrays for rankdata
      deallocate (tservalue,rankvalue,rankvaluemax,rankxmax,rankymax)

c --- Integer arrays for daccess
      deallocate (idatim)

c --- Real arrays for daccess
      deallocate (rdatim)


      return
      end
